""" 

command line interface for generating HEAL data dictionary/vlmd json files

"""
from functools import partial
import json
from pathlib import Path
import petl as etl
import pandas as pd
import csv
from collections import deque
import click
from slugify import slugify

from healdata_utils import mappings
from healdata_utils.transforms.excel.conversion import convert_dataexcel
from healdata_utils.transforms.csvdatadict.conversion import convert_datadictcsv
from healdata_utils.transforms.jsontemplate.conversion import convert_templatejson

from healdata_utils.transforms.stata.conversion import convert_stata
from healdata_utils.transforms.spss.conversion import convert_spss
from healdata_utils.transforms.sas.conversion import convert_sas
from healdata_utils.transforms.redcapcsv.conversion import convert_redcapcsv
from healdata_utils.transforms.csvdata.conversion import convert_datacsv
from healdata_utils.transforms.frictionless.conversion import (
    convert_frictionless_tableschema,
)

from healdata_utils.validators.validate import validate_vlmd_json, validate_vlmd_csv
from healdata_utils.utils import find_docstring_desc
from healdata_utils import schemas
# TODO: convert_templatecsv is misleading as it maps variable types. Need, to build this out further to support more translations.
# for now, just changing function name here.

choice_fxn = {
    "csv-version-update":partial(
        convert_datadictcsv,    
        renamemap=mappings.versions.fields_renamemap,
        recodemap=mappings.versions.fields_recodemap,
        droplist=mappings.versions.fields_droplist
    ),
    "excel-data":convert_dataexcel,
    "csv-data": convert_datacsv,
    'csv-data-dict':partial(
        convert_datadictcsv, 
        renamemap=mappings.renamemap,
        recodemap=mappings.recodemap
    ),
    #'csv-template':convert_templatecsv,
    "spss": convert_spss,
    "stata": convert_stata,
    #'por':convert_readstat,
    "sas": convert_sas,
    #'json-template':convert_templatejson,
    "redcap-csv": convert_redcapcsv,
    "frictionless-tbl-schema": convert_frictionless_tableschema,
}

# no extension for version updates (need to specify explicitly)
ext_map = {
    ".data.xlsx":"excel-data",
    ".datadict.csv":"csv-data-dict",
    ".data.csv":"csv-data",
    ".sav":"spss",
    ".sas7bdat":"sas",
    ".redcap.csv":'redcap-csv',
    ".dta":"stata"
}

# input_types = " - " + "\n - ".join(list(choice_fxn.keys()))

# input_descriptions = {
#     name: find_docstring_desc(fxn) for name, fxn in choice_fxn.items()
# }

input_short_descriptions = {
    "csv-version-update":"A data dictionary in a csv (or tsv file) that converts data dictionaries based on prior versions of the standard to the current version.",
    "excel-data":"Data (not metadata) in an excel (xlsx) workbook. Infers one data dictionary per sheet by default. Runs same inference as csv-data",
    "csv-data":"Data (not metadata) in a csv (or tsv) file. Infers variable/field name, type, and enum (possible values).",
    "csv-data-dict":"A minimal data dictionary in a csv (or tsv file). For example, may have name, type, and description and then adds the rest of the fields. Also maps a few common data types (such as char, character,text to 'string' and 'float' to 'number'.)",
    "spss":"A .sav file with data values,variable labels, and value labels. Extract labels and infers types from data.",
    "stata":"A .dta file with data values,variable labels, and value labels. Extract labels and infers types from data.",
    "sas":"A .sas7bdat file with data values and variable labels. Detects format (sas7bcat catalog files) in the same directory. Extract labels and infers types from data.",
    "redcap-csv":"A standardized REDCap csv data dictionary export file. Note, this must not include additional non-REDCap fields.",
    "frictionless-tbl-schema":"A frictionless table schema in json format."
}


def _write_vlmd(
    jsontemplate,
    csvtemplate,
    csvreport,
    jsonreport,
    output_filepath="heal-vlmd",
    output_overwrite=False,
    output_csv_quoting=None,
):
    # NOTE: currently the default is to auto generate file name if output_filepath not specified

    # flipped wording around for vars to avoid issues with previous flipping
    templatejson = jsontemplate
    templatecsv = csvtemplate
    reportjson = jsonreport
    reportcsv = csvreport


    output_filepath = Path(output_filepath)

    jsontemplate_path = output_filepath.with_suffix(".json")
    csvtemplate_path = output_filepath.with_suffix(".csv")

    # CHECK existence of directory and output files
    dir_exists = output_filepath.parent.exists()
    json_exists = jsontemplate_path.exists()
    csv_exists = csvtemplate_path.exists()

    if not dir_exists:
        raise NotADirectoryError(
            f"{str(output_filepath.parent)} does not exist so cannot create {output_filepath.name}"
        )

    if not output_overwrite:
        if json_exists and csv_exists:
            raise FileExistsError(
                f"{str(jsontemplate_path)}\nand\n{str(csvtemplate_path)}\nexist."
            )
        elif json_exists:
            raise FileExistsError(f"{str(jsontemplate_path)} exists.")
        elif csv_exists:
            raise FileExistsError(f"{str(csvtemplate_path)} exists.")


    if not reportcsv["valid"] or not reportjson["valid"]:
        reportdir = output_filepath.parent/"reports"
        reportdir.mkdir(exist_ok=True)

    # print JSON data dictionary and report
    jsontemplate_path.write_text(json.dumps(templatejson, indent=4))
    print()
    print(f"JSON data dictionary file written to {str(jsontemplate_path.resolve())}")

    if not reportjson["valid"]:
        reportjson_path = reportdir.joinpath("json-"+output_filepath.with_suffix(".json").name)
        reportjson_path.write_text(
            json.dumps(reportjson, indent=4)
        )
        print(f"but requires additional annotation and/or modifications.")
        print(f"View the report at: {str(reportjson_path.resolve())}")
        print()
    else:
        print()
    
    
    # print CSV data dictionary and report
    quoting = csv.QUOTE_NONNUMERIC if output_csv_quoting else csv.QUOTE_MINIMAL
    # NOTE: quoting non-numeric to allow special characters for nested delimiters within string columns (ie "=")
    # (
    #     etl.fromdicts(templatecsv)
    #     .tocsv(
    #         csvtemplate_path,
    #         quoting=csv.QUOTE_NONNUMERIC if output_csv_quoting else csv.QUOTE_MINIMAL)

    # )
    pd.DataFrame(templatecsv).to_csv(csvtemplate_path, quoting=quoting, index=False)
    print()
    print(f"CSV data dictionary file written to {str(csvtemplate_path.resolve())}")

    if not reportcsv["valid"]:
        reportcsv_path = reportdir.joinpath("csv-"+output_filepath.with_suffix(".json").name)
        reportcsv_path.write_text(
            json.dumps(reportcsv, indent=4)
        )
        print(f"but requires additional annotation and/or modifications.")
        print(f"View the report at: {str(reportcsv_path.resolve())}")
        print()
    else:
        print()


def _detect_inputtype(filepath,ext_to_inputtype=ext_map):        
    ext = "".join(input_filepath.suffixes)[1:].lower()
    inputtype = ext_to_inputtype.get(ext)

    if not inputtype:
        ext_to_inputtype_desc = "\n".join([ext+' for '+inputtype for ext,inputtype in ext_map.items()])
        raise Exception(
            f"No inputtype specified as file of type {ext} does not have a registered inputtype.", 
            "Either use the inputtype parameter or change your extensions to one of:",
            ext_to_inputtype_desc
        )
    return inputtype



def convert_to_vlmd(
    input_filepath,
    data_dictionary_props=None,
    inputtype=None,
    output_filepath=None,
    output_csv_quoting=None,
    output_overwrite=False,
    **kwargs,
):
    """
    Writes a data dictionary (i.e. variable level metadata) to a HEAL metadata JSON file using a registered function.

    Parameters
    ----------
    input_filepath : str
        Path to input file. See documentation on individual input types for more details.
    inputtype : str, optional
        The input type. If none specified, will default to using the file extension.
        See the currently registered input types in the input_types list.
    data_dictionary_props : dict, optional
        The other data-dictionary level properties. By default, will give the data_dictionary `title` property as the file name stem.
    output_filepath : str
        output file path or directory to where output will go. Note, the extension specified (csv or json will change as necessary)
    output_csv_quoting: bool, optional
        If true, all nonnumeric values will be quoted. This helps reduce ambiguity for programs
        like excel that uses special characters for specific purposes (eg = for formulas)
    output_overwrite: bool, optional
        If true, will overwrite files of the same name. 
    **kwargs: keyword arguments for specific registered input types
        currently this includes sas catalog file for sas and sheet_names/other params for excel.

    Returns
    -------
    dict
        Dictionary with:
         1. csvtemplated array of fields.
         2. jsontemplated data dictionary object as specified by an originally drafted design doc.
            That is, a dictionary with title:<title>,description:<description>,data_dictionary:<fields>
            where data dictionary is an array of fields as specified by the JSON schema.
         3. error objects for corresponding validators (ie frictionless for csv and jsonschema for json)
    NOTE
    ----
    In future versions, this will be more of a package bundled with corresponding schemas (whether csv or JSON),better organization
    (e.g., see frictionless Package).
    However, right now, it simply returns the csvtemplate and jsontemplate as specified
    in the heal specification repository.
    This is an intermediate solution to socialize a proof-of-concept.


    TODO
    --------

    Convert this to object-oriented framework -- 
    this may be best done for entire package (eg individual types as well)?

    This will require substantial overhaul to framework so this is not included in v.1 milestone.
    E.g., individual field types as objects, individual formats as objects
    
    Separate write and dd extraction -- this should be included in the OO revision above
    """

    input_filepath = Path(input_filepath)

    # infer input type
    inputtype = inputtype or _detect_inputtype(input_filepath)
    data_dictionary_props = data_dictionary_props or {}
        
    # ## add dd title
    # if not data_dictionary_props.get("title"):
    #     data_dictionary_props["title"] = input_filepath.stem

    # get data dictionary package based on the input type
    data_dictionary_package = choice_fxn[inputtype](
        input_filepath, data_dictionary_props,**kwargs
    )

    # with the need for multiple data dictionaries 
    # (eg within a package-like file resource such as excel or another archive format)
    # need to allow multiple data dictionaries. This was achieved by using the 
    # format similar to platform aggregate metadataservice disicovery page for 
    # data dictionaries: <key as name of dict>: ref to dictionary (or in this case - the actual data dictionary)
    # TODO: better way to identify one package of dds vs. multiple dds (as technically, templatejson/templatecsv could be keys in multiple dds)
    if "templatecsv" in data_dictionary_package and "templatejson" in data_dictionary_package:
        packages = {"":data_dictionary_package}
        onepackage = True
    else:
        packages = data_dictionary_package
        onepackage = False

    packages_with_reports = {}
    for name,package in packages.items():

        # add versions to both formats at top of object
        for field in package["templatecsv"]['fields']:
            field.update({"schemaVersion":schemas.healjsonschema["version"],**field})

        package["templatejson"] = {"schemaVersion":schemas.healjsonschema["version"],**dict(package["templatejson"])}
        
        package_csv = validate_vlmd_csv(
            package["templatecsv"]['fields'], to_sync_fields=True
        )
        package_json = validate_vlmd_json(package["templatejson"])

        # TODO: in future just return the packages (eg reports nested within package and not out of)
        # for now, keep same (report_xxx and templatexxx)

        reportcsv = package_csv["report"]
        reportjson = package_json["report"]
        dd_csv = package_csv["data"]
        dd_json = package_json["data"]


        if onepackage:

            packages_with_reports.update({
            "csvtemplate": dd_csv,
            "jsontemplate": dd_json,
            "errors": {"csvtemplate": reportcsv, "jsontemplate": reportjson},
            })

            if output_filepath:
                _write_vlmd(
                    jsontemplate=dd_json,
                    csvtemplate=dd_csv,
                    csvreport=reportcsv,
                    jsonreport=reportjson,
                    output_filepath=output_filepath,
                    output_overwrite=output_overwrite)
        else:

            packages_with_reports[name] = {
                "csvtemplate": dd_csv,
                "jsontemplate": dd_json,
                "errors": {"csvtemplate": reportcsv, "jsontemplate": reportjson},
            }

            # 

            if output_filepath:

                stem = Path(output_filepath).stem
                ext = Path(output_filepath).suffix
                output_filepath_with_name = Path(output_filepath).parent/(stem+"-"+slugify(name)+ext)

                _write_vlmd(
                    jsontemplate=dd_json,
                    csvtemplate=dd_csv,
                    csvreport=reportcsv,
                    jsonreport=reportjson,
                    output_filepath=output_filepath_with_name,
                    output_overwrite=output_overwrite
                )

                
    return packages_with_reports





from setuptools import setup, find_namespace_packages
from pathlib import Path
import os 

os.chdir(Path(__file__).parent)
def generate_long_description():
    return Path("README.md").read_text()

def get_install_requirements():
    return '''petl==1.7.12
jsonschema>=4.17.3
PyYaml>=6.0
pandas>=1.4
pyreadstat>=1.2.0
charset_normalizer>=2.1
visions>=0.7.5
click>=8.1.3
python-slugify
openpyxl

'''

setup(
    name='healdata_utils',
    version='0.6.0',
    author='Michael Kranz',
    author_email='kranz-michael@norc.org',
    long_description=generate_long_description(),
    long_description_content_type="text/markdown",    
    description='Data packaging tools for the HEAL data ecosystem',
    #TODO: change url to HEAL once migrated.
    url='https://github.com/norc-heal/healdata-utils',
    package_dir={'': 'src'},
    packages=find_namespace_packages(where='src'),
    install_requires=get_install_requirements(),
    entry_points='''
        [console_scripts]
        vlmd=healdata_utils.cli:vlmd
    '''

)



''' 

command line interface for generating HEAL data dictionary/vlmd json files

''' 

import click 
import json
from pathlib import Path
import petl as etl
import pandas as pd
import csv
from collections import deque

from healdata_utils.utils import find_docstring_desc
from healdata_utils.conversion import convert_to_vlmd,choice_fxn,input_short_descriptions
from healdata_utils.validators import validate_vlmd_json,validate_vlmd_csv
from healdata_utils.io import write_vlmd_template
from healdata_utils.config import VLMD_DEFS_URL

inputtype_descs = "\n".join([click.style(name,bold=True)+": "+desc for name,desc in input_short_descriptions.items()])

prompt_file = f"""
{click.style("Enter the path to your file:",bold=True,fg="green")}

This can be an:
1. absolute path (e.g., C:/Users/lastname-firstname/projectfolder/data/filename.csv)
2. relative path (e.g., data/filename.csv)
3. filename (e.g., filename.csv)

"""

prompt_extract_inputtypes = f"""
{click.style("What type of file do you want to extract variable level metadata from?",bold=True,fg="green")}

{inputtype_descs}

"""

prompt_template_nfields = f"""
{click.style("How many variables (ie fields) are in your data (ie are going to be in your data dictionary)?",bold=True,fg="green")}

""" 

prompt_template_outputfile = f"""

{click.style("What do you want the output file called?",bold=True,fg="green")}

This can be an:
1. absolute path (e.g., C:/Users/lastname-firstname/projectfolder/heal-data-dd)
2. relative path (e.g., heal-data-dd)
3. filename (e.g., heal-data-dd)

Note, if you specify a csv file, this will generate a HEAL csv templated file and if you specify a json file, this
will generate a HEAL json templated file.

""" 
prompt_extract_outputfile = f""" 

{click.style("What do you want the output file called?",bold=True,fg="green")}

Note, both a json and csv versions will be created.

"""

prompt_overwrite = f"""

{click.style("Do you want to overwrite the specified output files (if they exist)?",bold=True,fg="green")}

"""

def _check_suffix_for_json_or_csv(ctx,param,value):
    fileexts = [".json",".csv"]
    if not Path(value).suffix in fileexts:
        click.secho(f"File must be {' or '.join(fileexts)}",fg="red")
        click.pause(click.style("Press any key to exit program and try again"))
        ctx.exit()


def _check_overwrite(ctx,param,value):

    """ 
    Takes the specified file/outputfile argument 
    if the file exists, prompts user if they want to overwrite.

    Then, if the user has specified no on overwrite, the command 
    will abort. 


    # if no output file, then pass through (as not applicable)
    # if overwrite specified pass through (true)
    # if overwrite not specified (false), check existence and prompt user if they want to overwrite,
    #  if files don't exist, pass through (false)

    NOTE
    ----- 
    Note, both the csv and json files are checked for existence. 
    Reasoning for checking both csv and json is to ensure they are "synced" and
    currently the convert_to_vlmd fxn outputs both csv and json by default

    """
    
    filepath = ctx.params.get("outputfile")


    if not filepath or value:
        return value

    filepath_csv = Path(filepath).with_suffix(".csv")
    filepath_json = Path(filepath).with_suffix(".json")

    if filepath_json.exists() or filepath_csv.exists():

        # Write to console the existence of csv and json files
        # with specified stem
        if filepath_csv.exists():
            click.secho(f"Warning: {filepath_csv} exists",fg="red")

        if filepath_json.exists():
            click.secho(f"Warning: {filepath_json} exists",fg="red")  

        click.secho(f"Given you do not want to overwrite files and files exist, exiting tool.",fg="red")
        click.pause(click.style("Press any key to exit program and try again"))
        ctx.exit()
    
    return value        
                        


@click.group(invoke_without_command=True)
@click.pass_context
def vlmd(ctx):

    if ctx.invoked_subcommand is None:
        subcmds = []
        

        info = ctx.command.to_info_dict(ctx)["commands"]
        prompt_subcmds = f"{click.style('START BY TYPING ONE OF THE THREE COMMANDS:',underline=True)}\n\n"
        for subcmd_name,subcmd_info in info.items():

            subcmds.append(subcmd_name)
            subcmd_help = f"{click.style(subcmd_name,bold=True)}: "
            subcmd_help += f'{subcmd_info["help"]+":" if subcmd_info["help"] else ""}\n'
            prompt_subcmds += subcmd_help + "\n"

        subcmd = click.prompt(
            text=prompt_subcmds,
            type=click.Choice(subcmds)
        )
        # Determine if subcmd has an argument(s)
        has_args = False
        for p in ctx.command.commands[subcmd].params:
            if type(p) is click.core.Argument:
                has_args = True

        # NOTE: prompt for argument here (as prompts only for options)
        if has_args:
            filepath = click.prompt(
                text=prompt_file,
                type=click.Path()
            )
            # invoke subcommand chosen with "all bells and whistles"
            #NOTE: for Command.main method:
            # https://github.com/pallets/click/blob/b63ace28d50b53e74b5260f6cb357ccfe5560133/src/click/core.py#L1255
            ctx.command.commands[subcmd].main(args=[filepath])
        else:
            ctx.command.commands[subcmd].main()
            

# @vlmd.command(help="Lookup the definition and examples for a given field within a given data dictionary format or for both")
# @vlmd.option("--format",
#     default="both",
#     type=click.Choice(["csv","json","both"]),
#     help="Format to look up",
#     prompt="Which format do you want to look up?")
# @vlmd.option("--name",default="description",
# def lookup():
#     pass 

@vlmd.command(help="Extract the variable level metadata from an existing file with a specific type/format")
@click.argument("inputfile",type=click.Path(exists=True))
#TODO: --output-file or --output-filepath?
@click.option('--inputtype',type=click.Choice(list(choice_fxn.keys())),
    prompt=prompt_extract_inputtypes)
@click.option('--outputfile',
    default="heal-dd",
    help=inputtype_descs,
    prompt=prompt_extract_outputfile)
@click.option('--overwrite',default=False,is_flag=True,callback=_check_overwrite,
    help="If true, will replace (overwrite) the existing file if it exists. If false (the default) and there is a file of same name, will prompt user if they want to overwrite.")
@click.option(
    "--prop",
    "data_dictionary_props",
    multiple=True,nargs=2,
    help="<name of root level data dictionary property> <value of the specified property>. For example, title 'This is my title' Remember quotes if spaces!"
)
def extract(inputfile,outputfile,inputtype,overwrite,data_dictionary_props):

    data_dictionary_props = dict(data_dictionary_props)

   # data_dictionary_props = {}
    #save dds and error reports to files
    #NOTE: ported the sas catalog inference to core function
    data_dictionaries = convert_to_vlmd(
        input_filepath=inputfile,
        #data_dictionary_props=data_dictionary_props,
        output_filepath=outputfile,
        inputtype=inputtype,
        output_overwrite=overwrite,
        data_dictionary_props=data_dictionary_props
    )


@vlmd.command(help="Check (validate) an existing HEAL data dictionary file to see if it follows the HEAL specifications.")
@click.argument("inputfile",type=click.Path(exists=True))
@click.option('--outputfile',help="Write the report to a file. By default, the report will be printed directly to the console.")
@click.option('--overwrite',default=False,is_flag=True,callback=_check_overwrite)
def validate(inputfile,outputfile,overwrite):

    ext = Path(inputfile).suffix.replace(".","")

    if ext == "csv":
        package = validate_vlmd_csv(inputfile, to_sync_fields=True)
        report = package["report"]
    elif ext == "json":
        package = validate_vlmd_json(inputfile)
        report = package["report"]
    else:
        raise Exception("Need to specify either a csv or json file")

    if outputfile:
        Path(outputfile).write_text(
            json.dumps(report, indent=4)
        )
    else:
        #TODO: color code; more informative errors
        click.secho(json.dumps(report, indent=4))
         
@vlmd.command(help="Launch the vlmd data dictionary definitions in the documentation")
def documentation():
    click.launch(VLMD_DEFS_URL)

@vlmd.command(help="Start a data dictionary from an empty template")
@click.argument("outputfile",type=click.Path())
@click.option('--overwrite',default=False,is_flag=True,callback=_check_overwrite)
@click.option("--numfields",default=1,type=int,prompt=prompt_template_nfields,help="The number of fields (variables) in your data dictionary. This is used to make the template to write empty fields in the template.")
def template(outputfile,overwrite,numfields):
    write_vlmd_template(outputfile,output_overwrite=overwrite,numfields=numfields)

        
if __name__=='__main__':
    vlmd()


# """ 
# REDCAP input CSV schema
# """ 

# from frictionless import Schema 

# schema = Schema({
#   "fields": [
#     {
#       "type": "string",
#       "name": "Variable / Field Name"
#     },
#     {
#       "type": "string",
#       "name": "Form Name"
#     },
#     {
#       "type": "string",
#       "name": "Section Header"
#     },
#     {
#       "type": "string",
#       "name": "Field Type"
#     },
#     {
#       "type": "string",
#       "name": "Field Label"
#     },
#     {
#       "type": "string",
#       "name": "Choices, Calculations, OR Slider Labels"
#     },
#     {
#       "type": "string",
#       "name": "Field Note"
#     },
#     {
#       "type": "string",
#       "name": "Text Validation Type OR Show Slider Number"
#     },
#     {
#       "type": "number",
#       "name": "Text Validation Min"
#     },
#     {
#       "type": "number",
#       "name": "Text Validation Max"
#     },
#     {
#       "type": "string",
#       "name": "Identifier?"
#     },
#     {
#       "type": "string",
#       "name": "Branching Logic (Show field only if...)"
#     },
#     {
#       "type": "string",
#       "name": "Required Field?"
#     },
#     {
#       "name": "Custom Alignment",
#       "type": "any"
#     },
#     {
#       "name": "Question Number (surveys only)",
#       "type": "any"
#     },
#     {
#       "type": "string",
#       "name": "Matrix Group Name"
#     }
#   ]
# })




""" convert various frictionless objects to
another format """
from healdata_utils.transforms.jsontemplate.conversion import convert_templatejson
from healdata_utils import schemas
import os 
import yaml
import json
from pathlib import Path

def convert_frictionless_tableschema(
    schema,
    data_dictionary_props:dict=None,
    **kwargs
):
    if isinstance(schema,(os.PathLike,str)):
        if Path(schema).suffixes[-1]==".yaml": #NOTE: not tested
            schemajson = yaml.safe_load(Path(schema).read_text())
        else:
            schemajson = json.loads(Path(schema).read_text())
    else:
        schemajson = schema


    data_dictionaries = convert_templatejson(
        schemajson,
        data_dictionary_props,
        fields_name="fields",
        **kwargs
    )
    return data_dictionaries


def convert_frictionless_to_jsonschema(frictionless_schema):
    """ converts a frictionless schema to jsonschema """
    
    frictionless_fields = list(frictionless_schema.get("fields"))
    assert frictionless_fields,"A frictionless schema MUST have a set of fields"


    # schema level properties

    # frictionless
    missing_values = frictionless_schema.get("missingValues",[None,""])
    primary_keys = frictionless_schema.get("primaryKeys",[])

    # frictionless --> jsonschema per field
    jsonschema_properties = {}
    fieldnames = []
    for field in frictionless_fields:
        fieldnames.append(field["name"])
        jsonschema_properties[field["name"]] = prop = {}

        #base props
        if "description" in field:
            prop["description"] = field["description"]
        if "title" in field:
            prop["title"] = field["title"]
        if "example" in field:
            prop["example"] = field["example"]
        if "type" in field:
            if field["type"]!="any":
                prop["type"] = field["type"]
        
        

        # constraints
        constraints = field.get("constraints",{})
        if "enum" in constraints:
            type_mappings = {
                "integer":int,
                "number":float,
                "string":str
            }
            prop["enum"] = []
            for val in constraints["enum"]:
                if field.get("type","") in type_mappings:
                    coerced_val = type_mappings[field["type"]](val)
                    prop["enum"].append(coerced_val)
                else:
                    prop["enum"].append(val)


        if "pattern" in constraints:
            prop["pattern"] = constraints["pattern"] 

        if "minimum" in constraints:
            prop["minimum"] = constraints["minimum"]

        if "maximum" in constraints:
            prop["maximum"] = constraints["maximum"]


        # all fields are required - missing-ness vs. required in tabular perspective is tacked on to property
        is_required = "required" in constraints or field["name"] in primary_keys

        if is_required:
            # if required value: MUST be NOT missing value and the property
            prop_accounting_for_missing = {
                "allOf":[
                    prop,{"not": {"enum":missing_values}}
                ]
            }
        else:
            # if not required value: MUST be property OR the specified missing value
            prop_accounting_for_missing = {
                "anyOf":[
                    prop,{"enum":missing_values}
            ]}
        
        jsonschema_properties[field["name"]] = prop_accounting_for_missing
            

    
    jsonschema_schema = {
        "type":"object",
        "required":fieldnames, 
        "properties":jsonschema_properties
    }

    schema = frictionless_schema
    for jsonprop in ["description","title","name"]:
        if schema.get(jsonprop):
            jsonschema_schema[jsonprop] = schema[jsonprop]

    return {"type":"array","items":jsonschema_schema}